use crate::{gui::{VRAM_HEIGHT, VRAM_WIDTH}, MainWindow, VRAMImage};
use super::{GUIElementsRef, MainWindowRef};

use slint::{Model, SharedString};
use tim_tool::logic::tim::types::Encoding;
use std::{ops::RangeInclusive, rc::Rc, sync::{Arc, Mutex}};

struct VRAM {
    count:      usize,
    file_list:  Rc<slint::VecModel<slint::StandardListViewItem>>,
    image_list: Rc<slint::VecModel<VRAMImage>>
}

impl VRAM {
    pub fn push(&mut self, name: slint::StandardListViewItem, image: VRAMImage) {
        self.file_list.push(name);
        self.image_list.push(image);
        self.count += 1;
    }

    pub fn remove(&mut self, idx: usize) {
        self.count -= 1;
        self.file_list.remove(idx);
        self.image_list.remove(idx);
    }

    pub fn pop(&mut self) {
        self.remove(self.count - 1);
    }
}

pub struct MainTab {
    main_window:     MainWindowRef,
    vram:            Arc<Mutex<VRAM>>,
}

impl MainTab {
    pub fn new(main_window: MainWindowRef) -> MainTab {
        let vram_file_list:Vec<slint::StandardListViewItem> = main_window.borrow().get_main_tab_vram_file_list().iter().collect();
        let vram_file_list                                  = Rc::new(slint::VecModel::from(vram_file_list));
        let vram_image_list                                 = Rc::new(slint::VecModel::from(Vec::<VRAMImage>::new()));

        main_window.borrow().set_main_tab_vram_file_list(vram_file_list.clone().into());
        main_window.borrow().set_main_tab_vram_images(vram_image_list.clone().into());

        MainTab{main_window, vram: Arc::new(Mutex::new(VRAM{count: 0, file_list: vram_file_list, image_list: vram_image_list}))}
    }

    pub fn add_new_vram_file(&mut self, file_name: &String, full_image: slint::Image, image: slint::Image, encoding: Encoding, image_palette: Option<slint::Image>) -> usize {
        let mut vram_data     = self.vram.lock().expect("VRAM already locked");
        let mut add_new_image = |file_name: &String, full_image: slint::Image, image: slint::Image, encoding: Encoding, palette_count: i32, is_palette: bool| {
            let encoding_str = if is_palette {"<Palette>"} else {encoding.to_str()};
            let vram_image   = VRAMImage{
                full_img:      full_image,
                img:           image,
                x:             0,
                y:             0,
                encoding_str: SharedString::from(encoding_str),
                palette_count,
                is_palette,
            };

            vram_data.push(slint::StandardListViewItem::from(file_name.as_str()), vram_image);
        };

        let mut images_added = 1;
        add_new_image(file_name, full_image, image, encoding, if image_palette.is_some() {1} else {0}, false);
        if let Some(image_palette) = image_palette {
            let file_name = "  => ".to_owned() + file_name.as_str() + " (Palette)";

            images_added += 1;
            add_new_image(&file_name, image_palette.clone(), image_palette, encoding, 0, true);
        }

        images_added
    }

    pub fn remove_vram_file(&mut self, idx: usize) -> Result<RangeInclusive<usize>, &'static str> {
        let mut vram_data = self.vram.lock().expect("VRAM already locked");
        let     extras    = {
            if let Some(element) = vram_data.image_list.iter().skip(idx).next() {
                if element.is_palette {
                    return Err("Can not remove palette. Delete image instead");
                }

                else {
                    element.palette_count as usize
                }
            }

            else {
                0
            }
        };

        let     range        = idx..=idx+extras;
        let mut remove_image = |idx: usize| {
            vram_data.remove(idx);
        };

        for _ in range.clone() {
            remove_image(idx);
        }
        return Ok(range);
    }

    pub fn pop_vram_files(&mut self, count: usize) {
        let mut vram_data = self.vram.lock().expect("VRAM already locked");
        for _ in 0..count {
            vram_data.pop();
        }
    }

    pub fn move_vram_image(&mut self, idx: usize, dx: i32, dy: i32) {
        let vram_data = self.vram.lock().expect("VRAM already locked");
        if let Some(mut vram_info) = vram_data.image_list.row_data(idx) {
            vram_info.x += dx;
            vram_info.y += dy;

            if vram_info.x < 0 {
                vram_info.x = 0;
            }

            if vram_info.y < 0 {
                vram_info.y = 0;
            }

            let (vram_img_width, vram_img_height) = (vram_info.img.size().width as i32, vram_info.img.size().height as i32);
            if (vram_info.x + vram_img_width) > VRAM_WIDTH as i32 {
                vram_info.x = VRAM_WIDTH as i32 - vram_img_width;
            }

            if (vram_info.y + vram_img_height) > VRAM_HEIGHT as i32 {
                vram_info.y = VRAM_HEIGHT as i32 - vram_img_height;
            }

            vram_data.image_list.set_row_data(idx, vram_info);
        }
    }

    pub fn on_move_vram_image(&self, gui_elements: GUIElementsRef, mut function: impl FnMut(&mut MainTab, &MainWindow, i32, i32, i32) + 'static) {
        let main_window_cloned = self.main_window.clone();
        let gui_cloned         = gui_elements.clone();
        self.main_window.borrow().on_move_vram_image(move |idx, dx, dy| {
            function(&mut gui_cloned.borrow_mut().main_tab, &main_window_cloned.borrow(), idx, dx, dy);
        });        
    }

    pub fn on_remove_file(&self, gui_elements: GUIElementsRef, mut function: impl FnMut(&mut MainTab, &MainWindow, i32) + 'static) {
        let main_window_cloned = self.main_window.clone();
        let gui_cloned         = gui_elements.clone();
        self.main_window.borrow().on_main_tab_remove_file_clicked(move |idx| {
            function(&mut gui_cloned.borrow_mut().main_tab, &main_window_cloned.borrow(), idx);
        });
    }
}